
import 'dart:async';

import 'package:flutter/services.dart';

class FlutterItlBankDispenser {
  static const MethodChannel _channel = MethodChannel('flutter_itl_bank_dispenser');

  static const EventChannel _eventChannel = EventChannel('flutter_itl_bank_dispenser/flutter_plugin_callback');

  static const String getComPortsEvent = "getComPorts";
  static const String pollingCallBackEvent = "pollingCallBack";

  // Init default value;
  static const int baudRate = 9600;
  static const int timeout = 1000;
  static const int retryLevel = 3;

  // Empty Mode
  static const int normalEmptyMode = 1;
  static const int smartEmptyMode = 2;

  // Routing Mode
  static const int cashMode = 1;
  static const int smartPayMode = 2;

  static Future<String?> get platformVersion async {
    final String? version = await _channel.invokeMethod('getPlatformVersion');
    return version;
  }

  static Future<String?> get test async {
    final String? version = await _channel.invokeMethod('test');
    return version;
  }

  static Future<bool?>initialize(int baudRate,int timeout, int retryLevel) async {
    // command.BaudRate = 9600;
    // command.Timeout = 1000;
    // command.RetryLevel = 3;
    final params = <String, dynamic>{
      'baudRate': baudRate,
      'timeout': timeout,
      'retryLevel': retryLevel,
    };
    return await _channel.invokeMethod('initialize',params);
  }

  static Future<bool?>isInit() async {
    return await _channel.invokeMethod('isInit');
  }

  static Future<bool?> get getComPorts async {
    // final List<Map<dynamic, dynamic>>? comPorts = List<Map<dynamic, dynamic>>.from(await _channel.invokeMethod('getComPorts'));
    // return comPorts;

    return await _channel.invokeMethod('getComPorts');
  }

  static Future<Map<dynamic, dynamic>> connectComPort(int comPort)async{
    final params = <String, dynamic>{
      'comPort': comPort,
    };
    return await _channel.invokeMethod('connectComPort',params);
  }

  static Future<bool> disconnectComPort()async{
    return await _channel.invokeMethod('disconnectComPort');
  }

  static Future<bool> setPayoutStatus(bool isEnable)async{
    final params = <String, dynamic>{
      'isEnable': isEnable.toString(),
    };
    return await _channel.invokeMethod('setPayoutStatus',params);
  }

  static Future<bool> setPolling(bool isEnable)async{
    final params = <String, dynamic>{
      'isEnable': isEnable.toString(),
    };
    return await _channel.invokeMethod('setPolling',params);
  }

  // 1 - Normal Empty
  // 2 - Smart Empty
  static Future<bool?> empty(int type) async {
    // command.BaudRate = 9600;
    // command.Timeout = 1000;
    // command.RetryLevel = 3;
    final params = <String, dynamic>{
      'emptyMode': type,
    };
    return await _channel.invokeMethod('empty',params);
  }

  static Future<bool> setRouting(int routingType, int amount, String currency)async{
    final params = <String, dynamic>{
      'routingType': routingType,
      'amount': amount,
      'currency': currency,
    };
    return await _channel.invokeMethod('setRouting',params);
  }

  static Future<List<Map<dynamic, dynamic>>> getOutputLevelInfo()async{
    return List<Map<dynamic, dynamic>>.from(await _channel.invokeMethod('getOutputLevelInfo'));
  }

  static Future<List<Map<dynamic, dynamic>>> getNumChannel()async{
    return List<Map<dynamic, dynamic>>.from(await _channel.invokeMethod('getNumChannel'));
  }

  static Future<bool> disposePayout()async{
    return await _channel.invokeMethod('disposePayout');
  }

  static Stream<Map<dynamic, dynamic>> eventChannelStream() {
    return _eventChannel
        .receiveBroadcastStream()
        .map((result) {
      return Map<dynamic, dynamic>.from(result);
      // return List<Map<dynamic, dynamic>>.from(helloWorld);
    });
    // .distinct();
  }
}
