#include "include/flutter_itl_bank_dispenser/flutter_itl_bank_dispenser_plugin.h"

// This must be included before many other Windows headers.
#include <windows.h>

// For getPlatformVersion; remove unless needed for your plugin implementation.
#include <VersionHelpers.h>

#include <flutter/method_channel.h>
#include <flutter/plugin_registrar_windows.h>
#include <flutter/standard_method_codec.h>

#include <map>
#include <memory>
#include <sstream>
#include <flutter/event_channel.h>
#include <mutex>

#include <vector>
#include <time.h>
#include <thread>
#include "IO.h"
#include "CSMARTPayout.h"

using flutter::EncodableMap;
using flutter::EncodableValue;
using flutter::EncodableList;

namespace {

template<typename T = flutter::EncodableValue>
class MyStreamHandler: public flutter::StreamHandler<T> {
public:
	MyStreamHandler () = default;
	virtual ~MyStreamHandler () = default;

	void on_callback (flutter::EncodableValue _data) {
        std::unique_lock<std::mutex> _ul (m_mtx);
        if (m_sink.get ())
            m_sink.get ()->Success(_data);
    }

protected:
	std::unique_ptr<flutter::StreamHandlerError<T>> OnListenInternal (const T *arguments, std::unique_ptr<flutter::EventSink<T>> &&events) override {
        std::unique_lock<std::mutex> _ul (m_mtx);
		m_sink = std::move (events);
        return nullptr;
	}
	std::unique_ptr<flutter::StreamHandlerError<T>> OnCancelInternal (const T *arguments) override {
        std::unique_lock<std::mutex> _ul (m_mtx);
		m_sink.release ();
        return nullptr;
	}
private:
	flutter::EncodableValue m_value;
    std::mutex m_mtx;
	std::unique_ptr<flutter::EventSink<T>> m_sink;
};

std::vector<string> GetComPorts(); // Returns a vector containing a list of all the available ports on the host
bool ParseKeys(CPayout* payout); // The method responsible for parsing and responding to key presses
void DisplayCommands(); // Displays a list of commands available for this SDK
unsigned char FormatComPort(const string& comPort); // Remove the x. COM from the COM port string
DWORD WINAPI PollThread(LPVOID params); // The polling function, run as a seperate thread
HANDLE StartPollThread(CPayout* payout); // The method to start the polling thread, returns a handle to it
std::wstring ConvertTowstr(string stringtoconvert);
EncodableList GetComPortsResult(std::vector<string> ports);
EncodableMap WrapResultList(string type, EncodableValue list);
EncodableMap GetResultBody(bool result, string errorMsg, string returnMsg, string code);
void callBackEventChannel(string type, EncodableValue value);

const int PollInterval = 250; // Interval between polls (ms)
volatile bool IsPolling = false; // Indicates whether the poll thread should actively poll the unit
volatile bool PollFailed = false; // Indicates whether a poll has failed, can be called by multiple threads

// Payout Object
CPayout* Payout;
HANDLE payoutThread;
SSP_COMMAND command;

#define MAX_PROTOCOL_VERSION 8 // Max protocol version supported by SMART Payout (3/11/11).

class FlutterItlBankDispenserPlugin : public flutter::Plugin {
 public:
  static void RegisterWithRegistrar(flutter::PluginRegistrarWindows *registrar);

  FlutterItlBankDispenserPlugin();

  virtual ~FlutterItlBankDispenserPlugin();

 private:
  // Called when a method is called on this plugin's channel from Dart.
  void HandleMethodCall(
      const flutter::MethodCall<flutter::EncodableValue> &method_call,
      std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result);
};

MyStreamHandler<> *m_handler;
std::unique_ptr<flutter::EventChannel<flutter::EncodableValue>> m_event_channel;

// static
void FlutterItlBankDispenserPlugin::RegisterWithRegistrar(
    flutter::PluginRegistrarWindows *registrar) {
  auto channel =
      std::make_unique<flutter::MethodChannel<flutter::EncodableValue>>(
          registrar->messenger(), "flutter_itl_bank_dispenser",
          &flutter::StandardMethodCodec::GetInstance());

  auto plugin = std::make_unique<FlutterItlBankDispenserPlugin>();
    m_event_channel = std::make_unique<flutter::EventChannel<flutter::EncodableValue>> (
      registrar->messenger (), "flutter_itl_bank_dispenser/flutter_plugin_callback",
      &flutter::StandardMethodCodec::GetInstance()
    );

  channel->SetMethodCallHandler(
      [plugin_pointer = plugin.get()](const auto &call, auto result) {
        plugin_pointer->HandleMethodCall(call, std::move(result));
      });

    MyStreamHandler<> *_handler=new MyStreamHandler<> ();
    m_handler = _handler;
    auto _obj_stm_handle = static_cast<flutter::StreamHandler<flutter::EncodableValue>*> (m_handler);
    std::unique_ptr<flutter::StreamHandler<flutter::EncodableValue>> _ptr {_obj_stm_handle};
    m_event_channel->SetStreamHandler (std::move (_ptr));

  registrar->AddPlugin(std::move(plugin));
}

FlutterItlBankDispenserPlugin::FlutterItlBankDispenserPlugin() {}

FlutterItlBankDispenserPlugin::~FlutterItlBankDispenserPlugin() {}

void FlutterItlBankDispenserPlugin::HandleMethodCall(
    const flutter::MethodCall<flutter::EncodableValue> &method_call,
    std::unique_ptr<flutter::MethodResult<flutter::EncodableValue>> result) {
  if (method_call.method_name().compare("getPlatformVersion") == 0) {
    std::ostringstream version_stream;
    version_stream << "Windows ";
    if (IsWindows10OrGreater()) {
      version_stream << "10+";
    } else if (IsWindows8OrGreater()) {
      version_stream << "8";
    } else if (IsWindows7OrGreater()) {
      version_stream << "7";
    }
    result->Success(flutter::EncodableValue(version_stream.str()));
  }
  else if(method_call.method_name().compare("test") == 0){
      // The main class used to interface with the validator
      Payout = new CPayout();

      // Create a new thread here, this thread will not actively poll the unit until the
      // bool IsPolling is set to true.
      payoutThread = StartPollThread(Payout);
      // Set output to std::cout
      Payout->SetOutputStream(&std::cout);
      WriteString("C++ SDK\n");

      // Find out about the validator and create a connection struct to pass
      // to the validator class
      command;
      command.BaudRate = 9600;
      command.Timeout = 1000;
      command.RetryLevel = 3;
      command.IgnoreError = 1;

      // Get port number
      std::vector<string> ports = GetComPorts();
      WriteString(std::to_string(ports.size()).c_str());
      for (unsigned int i = 0; i < ports.size(); ++i){
          WriteString(ports[i]);
      }
      command.PortNumber = FormatComPort(ports[0]);
      command.SSPAddress = (unsigned char)0;
      // Get protocol version to use
      unsigned int p = 6;
      // Now connect to validator
      if (Payout->ConnectToPayout(command, p, 5) && Payout->GetUnitType() == 0x06)
      {
          WriteString("Connected to SMART Payout successfully");
          // Display the available commands
          DisplayCommands();

          Payout -> EnableValidator();

          IsPolling = true;
          WriteString("Started polling");
          // Begin checking input and poll fails
          while  (true)
          {
              if (!ParseKeys(Payout))
              {
                  IsPolling = false;
                  break; // If false break the loop
              }
              // This bool is set by the poll loop thread
              if (PollFailed)
              {
                  cout << "Poll failed" << endl;
                  if (Payout->ConnectToPayout(command, p, 5))
                  {
                      PollFailed = false;
                      StartPollThread(Payout);
                  }
                  else
                      break;
              }
          }
      }
      else if (Payout->GetUnitType() != 0x06)
          WriteString("Incorrect unit detected, this SDK supports the SMART Payout only");
      else
          WriteString("Failed to connect to the unit");

      cout << "Program shutting down... ";
      // Close poll thread
      TerminateThread(payoutThread, 1);
  	CloseHandle(payoutThread);
  	// Delete payout
  	delete Payout;

      result->Success(EncodableValue("123"));
    }
    else if(method_call.method_name().compare("initialize") == 0){
      try {
        if(Payout != nullptr && Payout->IsInit()){
          result->Success(EncodableValue(true));
          return;
        }

        WriteString("initialize\n");
        const auto* arguments = std::get_if<flutter::EncodableMap>(method_call.arguments());

        auto vBaudRate = arguments->find(flutter::EncodableValue("baudRate"));
        auto vTimeout = arguments->find(flutter::EncodableValue("timeout"));
        auto vRetryLevel = arguments->find(flutter::EncodableValue("retryLevel"));

        auto baudRate = vBaudRate != arguments->end() ? std::get<int>(vBaudRate->second) : 9600;
        auto timeout = vTimeout != arguments->end() ? std::get<int>(vTimeout->second) : 1000;
        auto retryLevel = vRetryLevel != arguments->end() ? std::get<int>(vRetryLevel->second) : 3;

        // The main class used to interface with the validator
        Payout = new CPayout();
        if(Payout == nullptr || !Payout->IsInit()){
          result->Success(EncodableValue(false));
          return;
        }

        // Create a new thread here, this thread will not actively poll the unit until the
        // bool IsPolling is set to true.
        payoutThread = StartPollThread(Payout);
        // Set output to std::cout
        Payout->SetOutputStream(&std::cout);
        WriteString("C++ SDK\n");

        // Find out about the validator and create a connection struct to pass
        // to the validator class
        command.BaudRate = baudRate;
        command.Timeout = timeout;
        command.RetryLevel = (unsigned char)retryLevel;
        command.IgnoreError = 1;
        command.SSPAddress = (unsigned char)0;

        result->Success(EncodableValue(true));
      }
      catch (...) {
        result->Success(EncodableValue(false));
      }
    }
    else if(method_call.method_name().compare("isInit") == 0){
        if(Payout == nullptr || !Payout->IsInit()){
          result->Success(EncodableValue(false));
        }else{
          result->Success(EncodableValue(true));
        }
      }
    else if(method_call.method_name().compare("getComPorts") == 0){
      // std::vector<string> ports = GetComPorts();

      thread t1(GetComPorts);
      t1.detach();
      // result->Success(GetComPortsResult(ports));
      // EncodableList out;
      // EncodableMap map;
      // map[EncodableValue("port")] = EncodableValue(1);
      // out.push_back(map);
      result->Success(EncodableValue(true));
    }
    else if(method_call.method_name().compare("disconnectComPort") == 0){
      if(Payout == nullptr && Payout->IsInit()){
        Payout->CloseComPort();
        result->Success(EncodableValue(true));
      }else{
        result->Success(EncodableValue(false));
      }

    }
    else if(method_call.method_name().compare("connectComPort") == 0){
        const auto* arguments = std::get_if<flutter::EncodableMap>(method_call.arguments());

        auto vComPort = arguments->find(flutter::EncodableValue("comPort"));
        auto comPort = vComPort != arguments->end() ? std::get<int>(vComPort->second) : 0;

        command.PortNumber = (unsigned char)comPort;

        // Result Params
        EncodableMap map;

        // Get protocol version to use
        unsigned int p = 6;
        // Now connect to validator
        if (Payout->ConnectToPayout(command, p, 5) && Payout->GetUnitType() == 0x06)
        {
          WriteString("Connected to SMART Payout successfully");
          map[EncodableValue("result")] = EncodableValue(true);
          map[EncodableValue("returnMsg")] = EncodableValue("Connected to SMART Payout successfully");
          map[EncodableValue("errorMsg")] = EncodableValue();
          result->Success(map);
        }
        else if (Payout->GetUnitType() != 0x06){
          WriteString("Incorrect unit detected, this SDK supports the SMART Payout only");
          map[EncodableValue("result")] = EncodableValue(false);
          map[EncodableValue("returnMsg")] = EncodableValue("");
          map[EncodableValue("errorMsg")] = EncodableValue("Incorrect unit detected, this SDK supports the SMART Payout only");
          result->Success(map);
        }
        else{
          WriteString("Failed to connect to the unit");
          map[EncodableValue("result")] = EncodableValue(false);
          map[EncodableValue("returnMsg")] = EncodableValue("");
          map[EncodableValue("errorMsg")] = EncodableValue("Failed to connect to the unit");
          result->Success(map);
        }
      }
      else if(method_call.method_name().compare("setPayoutStatus") == 0){
          const auto* arguments = std::get_if<flutter::EncodableMap>(method_call.arguments());
          auto vIsEnable = arguments->find(flutter::EncodableValue("isEnable"));
          auto isEnable = vIsEnable != arguments->end() ? std::get<string>(vIsEnable->second) : string{"false"};
          BOOL isEnableBool = isEnable == "true";

          EncodableMap map;
          if(isEnableBool){
              Payout -> EnableValidator();
              map = GetResultBody(true, "", "Enabled validator", "");
          }else{
              Payout -> DisableValidator();
              map = GetResultBody(true, "", "Disabled validator", "");
          }
          callBackEventChannel("pollingCallBack",map);
          result->Success(EncodableValue(true));
      }
      else if(method_call.method_name().compare("setPolling") == 0){
          const auto* arguments = std::get_if<flutter::EncodableMap>(method_call.arguments());
          auto vIsEnable = arguments->find(flutter::EncodableValue("isEnable"));
          auto isEnable = vIsEnable != arguments->end() ? std::get<string>(vIsEnable->second) : string{"false"};
          BOOL isEnableBool = isEnable == "true";

          if(isEnableBool){
              IsPolling = true;
              WriteString("Started polling");
          }else{
              IsPolling = false;
              WriteString("Stopped polling");
          }
          result->Success(EncodableValue(true));
      }
      else if(method_call.method_name().compare("empty") == 0){
        const auto* arguments = std::get_if<flutter::EncodableMap>(method_call.arguments());

        auto vEmptyMode = arguments->find(flutter::EncodableValue("emptyMode"));
        auto emptyMode = vEmptyMode != arguments->end() ? std::get<int>(vEmptyMode->second) : 0;

        if(emptyMode == 1){
          Payout->Empty();
          result->Success(EncodableValue(true));
        }
        else if(emptyMode == 2){
          Payout->SMARTEmpty();
          result->Success(EncodableValue(true));
        }
        else{
          result->Success(EncodableValue(false));
        }
      }
      else if(method_call.method_name().compare("getNumChannel") == 0){
          EncodableList list =  Payout->GetRecyclingStatus();

          EncodableMap resultMap;
          resultMap = GetResultBody(true, "", "GetNumChannel", "");
          resultMap[EncodableValue("data")] = list;
          callBackEventChannel("pollingCallBack",resultMap);
          result->Success(list);
      }
      else if(method_call.method_name().compare("getOutputLevelInfo") == 0){
          EncodableMap map;
          EncodableList list = Payout->GetOutputLevelInfo();

          map = GetResultBody(true, "", "GetOutputLevelInfo", "");
          map[EncodableValue("data")] = list;
          callBackEventChannel("pollingCallBack",map);
          result->Success(list);
      }
      else if(method_call.method_name().compare("setRouting") == 0){
        const auto* arguments = std::get_if<flutter::EncodableMap>(method_call.arguments());

        auto vRoutingType = arguments->find(flutter::EncodableValue("routingType"));
        auto vAmount = arguments->find(flutter::EncodableValue("amount"));
        auto vCurrency = arguments->find(flutter::EncodableValue("currency"));

        auto routingType = vRoutingType != arguments->end() ? std::get<int>(vRoutingType->second) : 1;
        auto amount = vAmount != arguments->end() ? std::get<int>(vAmount->second) : 0;
        auto currency = vCurrency != arguments->end() ? std::get<string>(vCurrency->second) : string{""};

        amount = amount *100;

        /// Route to CashBox
        if(routingType == 1){
          Payout->SendNoteToCashbox(amount, currency);
          result->Success(EncodableValue(true));
        }
        /// Route to SmartPayout
        else if(routingType == 2){
          Payout->SendNoteToStorage(amount, currency);
          result->Success(EncodableValue(true));
        }
        else{
          result->Success(EncodableValue(false));
        }
      }
      else if(method_call.method_name().compare("disposePayout") == 0){
          if(Payout->IsInit()){
              // Close poll thread
              if(payoutThread != null){
                  TerminateThread(payoutThread, 1);
                  CloseHandle(payoutThread);
              }

              Payout->CloseComPort();
              // Delete payout
              if(Payout != null){
                  delete Payout;
                  Payout = NULL;
              }
          }

          result->Success(EncodableValue(true));
      }
    else {
      result->NotImplemented();
    }
}

// This function searches through com ports 1 to 128 and determines if they
// exist or not. The list of existing ports is sent as a response.
EncodableList GetComPortsResult(std::vector<string> ports)
{
    EncodableList out;
    for(string port : ports){
        EncodableMap map;
        map[EncodableValue("port")] = EncodableValue(port);
        cout << port;
        out.push_back(map);
    }

    return out;
}

// This function searches through com ports 1 to 128 and determines if they
// exist or not. The list of existing ports is sent as a response.
std::vector<string> GetComPorts()
{
	vector<string> ports;
	// int index = 1;
	// Go through each possible port
	for (int i = 1; i <= 64; ++i)
    {
		char* buffer = new char[8];
		string port = "COM";
		port += std::to_string(i);
        WriteWString(ConvertTowstr(port));
		WriteString("\n");
		// First get the size
		DWORD dwSize = 0;
		LPCOMMCONFIG lpCC = (LPCOMMCONFIG) new BYTE[1];
		BOOL ret = GetDefaultCommConfig(ConvertTowstr(port).c_str(), lpCC, &dwSize);
		delete [] lpCC;
        WriteString(std::to_string(ret));

		// Now test the port
		lpCC = (LPCOMMCONFIG) new BYTE[dwSize];
		ret = GetDefaultCommConfig(ConvertTowstr(port).c_str(), lpCC, &dwSize);
        WriteString(std::to_string(ret));

		// If the port exists, add to the vector
		if (ret != null)
		{
			//string pusher = std::to_string(index++);
            string pusher = port;
			//pusher += ". ";
			//pusher += port;
			ports.push_back(pusher);
		}
		delete[] buffer;
		delete [] lpCC;
    }
    if (m_handler!=nullptr) {
        m_handler->on_callback(WrapResultList("getComPorts",GetComPortsResult(ports)));
    }
	return ports;
}

EncodableMap WrapResultList(string type, EncodableValue list){
    EncodableMap map;
    map[EncodableValue(type)] = list;
    return map;
}

EncodableMap GetResultBody(bool result, string errorMsg, string returnMsg, string code){
    EncodableMap map;
    map[EncodableValue("result")] = EncodableValue(result);
    map[EncodableValue("errorMsg")] = EncodableValue(errorMsg);
    map[EncodableValue("returnMsg")] = EncodableValue(returnMsg);
    map[EncodableValue("code")] = EncodableValue(code);

    return map;
}

void callBackEventChannel(string type, EncodableValue value){
    if (m_handler!=nullptr) {
        m_handler->on_callback(WrapResultList(type,value));
    }
}

std::wstring ConvertTowstr(string stringtoconvert){
    std::wstring temp = std::wstring(stringtoconvert.begin(), stringtoconvert.end());
	return temp;
}



bool ParseKeys(CPayout* payout)
{
	// If a keystroke is detected
	if (_kbhit())
	{
		// Parse the key
		switch (_getch())
		{
		case 'x': return false;
		case 'e':
			payout->EnableValidator();
			break;
		case 'd':
			payout->DisableValidator();
			break;
		case 'p':
			{
				// Toggle the polling
				if (IsPolling)
				{
					IsPolling = false;
					WriteString("Stopped polling");
				}
				else
				{
					IsPolling = true;
					WriteString("Started polling");
				}
				break;
			}
		case 'r':
			{
				// Stored notes
				WriteString("Notes stored:\n");
				payout->OutputLevelInfo();

				// Recycling status of notes
				WriteString("Recycling status of notes:\n");
				for (int i = 0; i < payout->GetNumChannels(); ++i)
				{
					cout << payout->GetUnitData()[i].Value*0.01f << " " << payout->GetUnitData()[i].Currency[0] <<
						payout->GetUnitData()[i].Currency[1] << payout->GetUnitData()[i].Currency[2] << " is ";
					(payout->GetUnitData()[i].Recycling)?cout << "being sent to storage" << endl:
						cout << "being sent to cashbox" << endl;
				}

				// Cashbox payout data
				WriteString("\nCashbox payout data:\n");
				payout->GetCashboxPayoutOpData();
				break;
			}
		case 'h': DisplayCommands(); break;
		case 't':
			{
				payout->ResetValidator();
				payout->CloseComPort(); // Force reconnection by closing com port
				break;
			}
		case 'a':
			{
				char choice = GetInputChar("1. Standard payout\n2. Payout by Denomination\r\n");
				if (choice == '1') // Standard payout
				{
					string s = GetInputString("Enter amount and currency to payout\nFormat is Amount Currency e.g. 10.00 EUR: ");

					// Break into 2 strings, first contains value, second currency
					string s1 = "";
					size_t pos = s.find_first_of(' ');
					if (pos == s.npos)
					{
						WriteString("Input was not in the correct format");
						break;
					}
					s1 = s.substr(pos+1, s1.npos);
					s.erase(pos, s.npos);

					// Convert value to integer
					float value = (float)atof(s.c_str());

					// Make sure each char in the currency is uppercase and
					// only 3 chars were entered
					if (s1.length() != 3)
					{
						WriteString("Invalid currency");
						break;
					}
					for (int i = 0; i < 3; ++i)
						s1[i] = (unsigned char)toupper(s1[i]);

					// If the value entered isn't 0 or negative, make the payout
					if (value > 0)
					{
						payout->Payout((int)(value * 100), s1);
					}
				}
				else if (choice == '2') // Payout by Denomination
				{
					// Get the dataset info from the validator
					SUnitData* UnitData = payout->GetUnitData();
					// Create a temporary array to hold denomination info
					unsigned char* dataArray = new unsigned char[255];
					int counter = 0, temp = 0, numDenoms = 0;
					// Go through each channel and query the user on how many of each denomination
					// they want to payout.
					for (int i = 0; i < payout->GetNumChannels(); ++i)
					{
						cout << "Number of " << UnitData[i].Value*0.01f << " " << UnitData[i].Currency[0] <<
							UnitData[i].Currency[1] << UnitData[i].Currency[2] << " to payout: ";
						temp = atoi(GetInputString("").c_str());
						if (temp != 0) // if they have entered a number
						{
							numDenoms++; // increase number of denominations to payout
							memcpy(dataArray + counter, &temp, 2); // copy number of denominations to the data array
							counter += 2;
							memcpy(dataArray + counter, &UnitData[i].Value, 4); // copy value of denomination to data array
							counter += 4;
							// copy currency of denom to data array
							dataArray[counter++] = UnitData[i].Currency[0];
							dataArray[counter++] = UnitData[i].Currency[1];
							dataArray[counter++] = UnitData[i].Currency[2];
						}
					}

					if (numDenoms > 0) // if the user didn't enter 0 for all denominations
					{
						payout->PayoutByDenomination(numDenoms, dataArray);
					}
					delete[] dataArray; // delete temp array once command sent
				}
				else
					WriteString("Invalid input\r\n");
				break;
			}
		case 'm':
			{
				// Empty has two choices, SMART or just empty. SMART keeps track of
				// what notes were emptied, empty does not.
				char choice = GetInputChar("1. Empty\n2. SMART Empty\n");
				if (choice == '1')
					payout->Empty();
				else if (choice == '2')
					payout->SMARTEmpty();
				else
					WriteString("Invalid input");
				break;
			}
		case 'c':
			{
				// Note recycling
				string s = GetInputString("Enter amount and currency to alter recycling on\nFormat is Amount Currency e.g. 10.00 EUR: ");

				// Break into 2 strings, first contains value, second currency
				string s1 = "";
				size_t pos = s.find_first_of(' ');
				if (pos == s.npos)
				{
					WriteString("Input was not in the correct format");
					break;
				}
				s1 = s.substr(pos+1, s1.npos);
				s.erase(pos, s.npos);

				// Convert value to integer
				float value = (float)atof(s.c_str());

				// Make sure each char in the currency is uppercase and
				// only 3 chars were entered
				if (s1.length() != 3)
				{
					WriteString("Invalid currency");
					break;
				}
				for (int i = 0; i < 3; ++i)
					s1[i] = (unsigned char)toupper(s1[i]);

				char c = GetInputChar("1. Store note\n2. Stack note\r\n");

				// Validation
				if ((c != '1' && c != '2'))
				{
					cout << "Invalid input" << endl;
					break;
				}

				// Convert to bool
				bool stack = atoi(&c)-1;

				// Convert value to penny value
				value *= 100;

				// Send to either cashbox or storage depending on choice
				(stack)?payout->SendNoteToCashbox((int)value, s1):payout->SendNoteToStorage((int)value, s1);

				break;
			}
		default: break;
		}
	}
	return true;
}

void DisplayCommands()
{
	string s = "Command List\n";
	s += "e = enable\nd = disable\n";
	s += "p = start/stop polling\n";
	s += "a = make a payout\n";
	s += "m = empty or SMART empty the unit\n";
	s += "c = change note recycling options\n";
	s += "t = reset validator\n";
	s += "r = report\n\n";
	s += "h = display this list again\n";
	s += "x = exit\n";
	WriteString(s);
}

// Remove the "x. COM" from the com port string
unsigned inline char FormatComPort(const string& portstring)
{
	string s = portstring;
	s.erase(0, 6);
	return (unsigned char)atoi(s.c_str());
}

// The following section contains threading functions. Threading is used in this program
// as it is vital that the validator is polled at regular intervals. Obviously the execution
// of the main thread of the program needs to be halted occasionally - such as when waiting
// for a user's input, so the polling of the validator needs to occur on a seperate thread
// which will never be halted. The validator class looks after critical sections and thread
// safing.

// This function is called to create a new instance of the poll thread. The function returns a
// handle to the thread.
HANDLE StartPollThread(CPayout* Payout1)
{
	unsigned int hopThreadID;
	return CreateThread(NULL,
						0,
						PollThread,
						Payout1,
						NULL,
						(LPDWORD)&hopThreadID);
}

// This function is responsible for polling the validator at a regular interval.
// It should be run as a seperate thread to the main execution thread.
DWORD WINAPI PollThread(LPVOID params)
{
	// Get validator instance from param
	CPayout* payout = reinterpret_cast<CPayout*>(params);

	// Keep looping until broken out of
	while (true)
	{
		// If the validator is actively being polled
		if (IsPolling)
		{
			// Attempt to send the poll
			if (!payout->DoPoll())
			{
				// If it fails (command can't get to validator) then set the failed poll
				// flag to true and exit the function. This will terminate the thread.
				PollFailed = true;
				return 0;
			}
			else {
                // parse poll response
                char* noteCurrency;
                // Result Params
                EncodableMap map;
                string callBackEventTitle = "pollingCallBack";

                for (int i = 1; i < payout->m_PollResponseLength; ++i)
                {
                    switch (payout->m_PollResponse[i])
                    {
                    // The unit is disabled and unable to accept notes or make payouts
                    case SSP_POLL_DISABLED:
                        *payout->m_Output << "Unit disabled" << endl;
                        map = GetResultBody(true,"","Unit disabled","0xE8");
                        callBackEventChannel(callBackEventTitle,map);

                        break;
                    // The unit has accepted a note as valid currency
                    case SSP_POLL_CREDIT:
                    {
                        float f = payout->GetChannelValue(payout->m_PollResponse[i + 1]) * 0.01f;
                        noteCurrency = payout->GetChannelCurrency(payout->m_PollResponse[i + 1]);
                        *payout->m_Output << "Credit " << f << " " << noteCurrency[0] <<
                            noteCurrency[1] << noteCurrency[2] << endl;
                        ++i;
                        string channelValue = std::to_string(f);
                        string channelCurrency;
                        channelCurrency.push_back(noteCurrency[0]);
                        channelCurrency.push_back(noteCurrency[1]);
                        channelCurrency.push_back(noteCurrency[2]);

                        string content = "Credit "+ channelValue + " " + channelCurrency;
                        map = GetResultBody(true,"",content,"0xEE");
                        map[EncodableValue("amount")] = EncodableValue(f);
                        map[EncodableValue("currency")] = EncodableValue(channelCurrency);
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                    }
                    // A note is being read, if the byte immediately following the response
                    // is 0 then the note has not finished reading, if it is greater than 0
                    // then this is the channel of that note.
                    case SSP_POLL_NOTE_READ:
                    {
                        if (payout->m_PollResponse[i + 1] > 0)
                        {
                            float f = payout->GetChannelValue(payout->m_PollResponse[i + 1]) * 0.01f;
                            noteCurrency = payout->GetChannelCurrency(payout->m_PollResponse[i + 1]);
                            *payout->m_Output << "Note in escrow, amount: ";
                            *payout->m_Output << f << " " << noteCurrency[0] <<
                                noteCurrency[1] << noteCurrency[2] << endl;

                            string channelValue = std::to_string(f);
                            string channelCurrency;
                            channelCurrency.push_back(noteCurrency[0]);
                            channelCurrency.push_back(noteCurrency[1]);
                            channelCurrency.push_back(noteCurrency[2]);

                            string content = "Note in escrow, amount: "+ channelValue + " " + channelCurrency;
                            map = GetResultBody(true,"",content,"0xEF");
                            map[EncodableValue("amount")] = EncodableValue(f);
                            map[EncodableValue("currency")] = EncodableValue(channelCurrency);
                            callBackEventChannel(callBackEventTitle,map);

                        }
                        else{
                            *payout->m_Output << "Reading note..." << endl;
                            map = GetResultBody(true,"","Reading note...","0xEF");
                            callBackEventChannel(callBackEventTitle,map);
                        }
                        ++i;
                        break;
                    }
                    // The unit is either stacking the note to the cashbox or to storage
                    case SSP_POLL_STACKING:
                        *payout->m_Output << "Stacking note..." << endl;
                        map = GetResultBody(true,"","Stacking note...","0xCC");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit has stacked the note in the cashbox
                    case SSP_POLL_STACKED:
                        *payout->m_Output << "Note stacked" << endl;
                        map = GetResultBody(true,"","Note stacked","0xEB");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit is in the process of rejecting a note
                    case SSP_POLL_REJECTING:
                        *payout->m_Output << "Note rejecting..." << endl;
                        map = GetResultBody(true,"","Note rejecting...","0xED");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit has rejected a note
                    case SSP_POLL_REJECTED:
                        *payout->m_Output << "Note rejected" << endl;
                        WriteString(payout->GetQueryRejectionErrorCode());
                        map = GetResultBody(false,payout->GetQueryRejectionErrorCode(),"Note rejecting","0xEC");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The cashbox has reached maximum capacity
                    case SSP_POLL_STACKER_FULL:
                        *payout->m_Output << "Stacker full" << endl;
                        map = GetResultBody(true,"","Stacker full","0xE7");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The cashbox is not attached to the unit
                    case SSP_POLL_CASHBOX_REMOVED:
                        *payout->m_Output << "Cashbox removed..." << endl;
                        map = GetResultBody(true,"","Cashbox removed...","0xE3");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The cashbox has been refitted to the unit
                    case SSP_POLL_CASHBOX_REPLACED:
                        *payout->m_Output << "Cashbox replaced" << endl;
                        map = GetResultBody(true,"","Cashbox replaced","0xE4");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // A fraud attempt has been detected
                    case SSP_POLL_FRAUD_ATTEMPT:
                        *payout->m_Output << "Fraud attempt!" << endl;
                        ++i;
                        map = GetResultBody(true,"","Fraud attempt!","0xE6");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // A note was detected inside the unit at startup and dispensed from the
                        // front of the unit
                    case SSP_POLL_NOTE_CLEARED_FROM_FRONT:
                    {
                        float f = payout->GetChannelValue(payout->m_PollResponse[i + 1]) * 0.01f;
                        noteCurrency = payout->GetChannelCurrency(payout->m_PollResponse[i + 1]);
                        *payout->m_Output << f << " " << noteCurrency[0] <<
                            noteCurrency[1] << noteCurrency[2] << " cleared from front" << endl;
                        ++i;

                        string channelValue = std::to_string(f);
                        string channelCurrency;
                        channelCurrency.push_back(noteCurrency[0]);
                        channelCurrency.push_back(noteCurrency[1]);
                        channelCurrency.push_back(noteCurrency[2]);

                        string content = channelValue + " " + channelCurrency + "  cleared from front";
                        map = GetResultBody(true,"",content,"0xE1");
                        map[EncodableValue("amount")] = EncodableValue(f);
                        map[EncodableValue("currency")] = EncodableValue(channelCurrency);
                        callBackEventChannel(callBackEventTitle,map);

                        break;
                    }
                    // A note was detected inside the unit at startup and stacked to the
                    // cashbox
                    case SSP_POLL_NOTE_CLEARED_TO_CASHBOX:
                    {
                        float f = payout->GetChannelValue(payout->m_PollResponse[i + 1]) * 0.01f;
                        noteCurrency = payout->GetChannelCurrency(payout->m_PollResponse[i + 1]);
                        *payout->m_Output << f << " " << noteCurrency[0] <<
                            noteCurrency[1] << noteCurrency[2] << " cleared to cashbox" << endl;
                        ++i;

                        string channelValue = std::to_string(f);
                        string channelCurrency;
                        channelCurrency.push_back(noteCurrency[0]);
                        channelCurrency.push_back(noteCurrency[1]);
                        channelCurrency.push_back(noteCurrency[2]);

                        string content = channelValue + " " + channelCurrency + " cleared to cashbox";
                        map = GetResultBody(true,"",content,"0xE2");
                        map[EncodableValue("amount")] = EncodableValue(f);
                        map[EncodableValue("currency")] = EncodableValue(channelCurrency);
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                    }
                    // A note has been sent to storage for payout
                    case SSP_POLL_NOTE_STORED_IN_PAYOUT:
                        *payout->m_Output << "Note stored for payout" << endl;
                        map = GetResultBody(true,"","Note stored for payout","0xDB");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit has been reset since the last time a poll was sent
                    case SSP_POLL_RESET:
                        *payout->m_Output << "Unit reset" << endl;
                        map = GetResultBody(true,"","Unit reset","0xF1");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit is in the process of dispensing a note
                    case SSP_POLL_DISPENSING:
                    {
                        string channelValue;
                        string channelCurrency;
                        string content;

                        int j = i + 2, l = payout->m_PollResponse[i + 1] * 7 + i + 2;
                        while (j < l)
                        {
                            int n = 0;
                            memcpy(&n, payout->m_PollResponse + (i + 2), 4);
                            j += 4;

                            unsigned char str1 = payout->m_PollResponse[j++];
                            unsigned char str2 = payout->m_PollResponse[j++];
                            unsigned char str3 = payout->m_PollResponse[j++];

                            *payout->m_Output << "Dispensing " << n * 0.01f << " ";
                            *payout->m_Output << str1;
                            *payout->m_Output << str2;
                            *payout->m_Output << str3 << "..." << endl;

                            channelValue = std::to_string(n * 0.01f);
                            channelCurrency = "";
                            channelCurrency.push_back(str1);
                            channelCurrency.push_back(str2);
                            channelCurrency.push_back(str3);

                            content = "Dispensing " + channelValue + " " + channelCurrency + "...";
                            map = GetResultBody(true, "", content, "0xDA");
                            map[EncodableValue("amount")] = EncodableValue(n * 0.01f);
                            map[EncodableValue("currency")] = EncodableValue(channelCurrency);
                            callBackEventChannel(callBackEventTitle,map);
                        }
                        i += l - 1;
                        break;
                    }
                    // The unit has finished dispensing a note
                    case SSP_POLL_DISPENSED:
                    {
                        string channelValue;
                        string channelCurrency;
                        string content;

                        int j = i + 2, l = payout->m_PollResponse[i + 1] * 7 + i + 2;
                        while (j < l)
                        {
                            unsigned char str1 = payout->m_PollResponse[j++];
                            unsigned char str2 = payout->m_PollResponse[j++];
                            unsigned char str3 = payout->m_PollResponse[j++];

                            int n = 0;
                            memcpy(&n, payout->m_PollResponse + (i + 2), 4);
                            j += 4;
                            *payout->m_Output << "Dispensed " << n * 0.01f << " ";
                            *payout->m_Output << str1;
                            *payout->m_Output << str2;
                            *payout->m_Output << str3 << endl;

                            channelValue = std::to_string(n * 0.01f);
                            channelCurrency = "";
                            channelCurrency.push_back(str1);
                            channelCurrency.push_back(str2);
                            channelCurrency.push_back(str3);

                            content = "Dispensed " + channelValue + " " + channelCurrency + "";
                            map = GetResultBody(true, "", content, "0xD2");
                            map[EncodableValue("amount")] = EncodableValue(n * 0.01f);
                            map[EncodableValue("currency")] = EncodableValue(channelCurrency);
                            callBackEventChannel(callBackEventTitle,map);
                        }
                        i += l - 1;
                        payout->EnableValidator();
                        break;
                    }
                    // The payout device is empty
                    case SSP_POLL_EMPTY:
                        *payout->m_Output << "Unit empty" << endl;
                        map = GetResultBody(true, "", "Unit empty", "0xD4");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit has become jammed
                    case SSP_POLL_JAMMED:
                        *payout->m_Output << "Unit jammed..." << endl;
                        map = GetResultBody(true, "", "Unit jammed...", "0xD4");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // A note has jammed in a place where the user may be able to retrieve it
                    case SSP_POLL_UNSAFE_JAM:
                        *payout->m_Output << "Unsafe jam detected..." << endl;
                        map = GetResultBody(true, "", "Unsafe jam detected...", "0xE9");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // A note has jammed in a place where the user cannot retrieve it
                    case SSP_POLL_SAFE_JAM:
                        *payout->m_Output << "Safe jam detected..." << endl;
                        map = GetResultBody(true, "", "Safe jam detected...", "0xEA");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit was interrupted while making a payout, some notes may have been dispensed
                    case SSP_POLL_INCOMPLETE_PAYOUT:
                    {
                        string channelValue;
                        string channelCurrency;
                        string content;

                        int j = i + 2;
                        int l = payout->m_PollResponse[i + 1] * 11 + j;
                        while (j < l)
                        {
                            int n = 0, n1 = 0;
                            memcpy(&n, payout->m_PollResponse + (i + 2), 4); // copy dispensed value
                            memcpy(&n1, payout->m_PollResponse + (i + 6), 4); // copy requested value
                            // Get currency
                            j += 8;
                            string curr = " ";

                            unsigned char str1 = payout->m_PollResponse[j++];
                            unsigned char str2 = payout->m_PollResponse[j++];
                            unsigned char str3 = payout->m_PollResponse[j++];

                            curr += str1;
                            curr += str2;
                            curr += str3;

                            *payout->m_Output << "Incomplete payout\nDispensed: ";
                            *payout->m_Output << n * 0.01f << curr << endl;
                            *payout->m_Output << "Requested: " << n1 * 0.01f << curr << endl;

                            channelValue = std::to_string(n * 0.01f);
                            channelCurrency = "";
                            channelCurrency.push_back(str1);
                            channelCurrency.push_back(str2);
                            channelCurrency.push_back(str3);

                            content = "Incomplete payout\nDispensed:  " + channelValue + " " + channelCurrency + "\n";
                            content += "Requested:  " + channelValue + " " + channelCurrency + "";

                            map = GetResultBody(true, "", content, "0xDC");
                            map[EncodableValue("amount")] = EncodableValue(n * 0.01f);
                            map[EncodableValue("currency")] = EncodableValue(channelCurrency);
                            callBackEventChannel(callBackEventTitle,map);

                        }
                        i += l - 1;
                        break;
                    }
                    // The unit is in the process of emptying its stored notes to the cashbox
                    case SSP_POLL_EMPTYING:
                        *payout->m_Output << "Emptying..." << endl;
                        map = GetResultBody(true, "", "Emptying...", "0xC2");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit has finished emptying its stored notes to the cashbox
                    case SSP_POLL_EMPTIED:
                        *payout->m_Output << "Unit emptied" << endl;
                        map = GetResultBody(true, "", "Unit emptied", "0xC3");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit is in the process of emptying its stored notes to the cashbox and keeping
                        // a track of what notes were emptied
                    case SSP_POLL_SMART_EMPTYING:
                        *payout->m_Output << "SMART emptying..." << endl;
                         map = GetResultBody(true, "", "SMART emptying...", "0xB3");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit has finished emptying its stored notes to the cashbox and has kept track of
                        // what was emptied. This data can be retrieved using the GetCashboxPayoutOpData method.
                    case SSP_POLL_SMART_EMPTIED:
                        *payout->m_Output << "SMART emptied" << endl;
                        payout->GetCashboxPayoutOpData(); // Report what was moved in the empty
                        map = GetResultBody(true, "", "SMART emptied", "0xB3");
                        map[EncodableValue("data")] = payout->GetCashboxPayoutOpDataList();
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // A note is resting in the bezel waiting to be removed by the user
                    case SSP_POLL_NOTE_IN_BEZEL:
                    {
                        int n = 0;
                        memcpy(&n, payout->m_PollResponse + (i + 1), 4);
                        *payout->m_Output << n * 0.01f << " ";
                        *payout->m_Output << payout->m_PollResponse[i + 5];
                        *payout->m_Output << payout->m_PollResponse[i + 6];
                        *payout->m_Output << payout->m_PollResponse[i + 7];
                        *payout->m_Output << " held in bezel" << endl;
                        i += 7;

                        EncodableMap data;
                        string channelCurrency;
                        channelCurrency += payout->m_PollResponse[i + 5];
                        channelCurrency += payout->m_PollResponse[i + 6];
                        channelCurrency += payout->m_PollResponse[i + 7];

                        map = GetResultBody(true, "", "Held in bezel", "0xCE");
                        data[EncodableValue("amount")] = EncodableValue(n * 0.01f);
                        data[EncodableValue("currency")] = EncodableValue(channelCurrency);
                        map[EncodableValue("data")] = data;
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                    }
                    // The unit has been opened exposing the note path
                    case SSP_POLL_NOTE_PATH_OPEN:
                        *payout->m_Output << "The note path is open" << endl;
                        map = GetResultBody(true, "", "The note path is open", "0xE0");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // The unit is disabled as all the channels are inhibited
                    case SSP_POLL_CHANNEL_DISABLE:
                        *payout->m_Output << "Unit disabled (all channels inhibited)" << endl;
                        map = GetResultBody(true, "", "Unit disabled (all channels inhibited)", "0xB5");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                        // Any poll response that is not detected is output with a warning
                    default:
                        *payout->m_Output << "WARNING: Unknown poll response detected: " << (int)payout->m_PollResponse[i] << endl;
                        map = GetResultBody(true, "", "WARNING: Unknown poll response detected: "+to_string((int)payout->m_PollResponse[i]), "");
                        callBackEventChannel(callBackEventTitle,map);
                        break;
                    }
                }
            }
			Sleep(PollInterval); // Wait for the interval
		}
	}
	return 0;
}

}  // namespace

void FlutterItlBankDispenserPluginRegisterWithRegistrar(
    FlutterDesktopPluginRegistrarRef registrar) {
  FlutterItlBankDispenserPlugin::RegisterWithRegistrar(
      flutter::PluginRegistrarManager::GetInstance()
          ->GetRegistrar<flutter::PluginRegistrarWindows>(registrar));
}
